/*
 *	linux/fs/rbf/namei.c
 *	RBF File System - name operations
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *
 * from
 *  
 *  linux/fs/ext2/namei.c
 *
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 * from
 *
 *  linux/fs/minix/namei.c
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include "rbf_page.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

/*
 * Couple of helper functions - make the code slightly cleaner.
 */

static inline void rbf_inc_count(struct inode *inode)
{
	inode_inc_link_count(inode);
	mark_inode_dirty(inode);
}

static inline void rbf_dec_count(struct inode *inode)
{
	inode_dec_link_count(inode);
	mark_inode_dirty(inode);
}

static inline int rbf_add_nondir(struct dentry *dentry, struct inode *inode)
{
	int err = rbf_add_link(dentry, inode);
	if (!err) {
		d_instantiate(dentry, inode);
		return 0;
	}
	rbf_dec_count(inode);
	iput(inode);
	return err;
}

/*
 * Methods themselves.
 */

static struct dentry *rbf_lookup(struct inode *dir, struct dentry *dentry, struct nameidata *nd)
{
	struct inode *inode;
	ino_t ino;
	
	if (dentry->d_name.len > RBF_NAMELEN)
		return ERR_PTR(-ENAMETOOLONG);

	lock_kernel();
	ino = rbf_inode_by_name(dir, dentry);
	inode = NULL;
	if (ino) {
		inode = iget(dir->i_sb, ino);
		if (!inode) {
			unlock_kernel();
			return ERR_PTR(-EACCES);
		}
	}
	unlock_kernel();
	d_add(dentry, inode);
	return NULL;
}

/*
 * By the time this is called, we already have created
 * the directory cache entry for the new file, but it
 * is so far negative - it has no inode.
 *
 * If the create succeeds, we fill in the inode information
 * with d_instantiate(). 
 */
static int rbf_create(struct inode *dir, struct dentry *dentry, int mode, struct nameidata *nd)
{
	struct super_block *s;
	struct inode *inode;
	int err;

	if (dir == NULL)
		return -ENOENT;

	s = dir->i_sb;
	inode = rbf_new_inode(dir, mode);
	err = PTR_ERR(inode);

	if (!IS_ERR(inode)) {
		inode->i_op = s_su_operations->file_inode_ops;
		inode->i_fop = s_su_operations->file_operations;
		inode->i_mapping->a_ops = s_su_operations->address_space_ops;
		mark_inode_dirty(inode);
		err = rbf_add_nondir(dentry, inode);
	}
	return err;
}


static int rbf_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	struct super_block *s = dir->i_sb;
	struct inode *inode = old_dentry->d_inode;

	if (S_ISDIR(inode->i_mode))
		return -EPERM;

	if (!s_su_hardlinks)
		return -EPERM;

	if (inode->i_nlink >= RBF_LINK_MAX)
		return -EMLINK;
	
	inode->i_ctime = CURRENT_TIME;
	rbf_inc_count(inode);
	atomic_inc(&inode->i_count);

	return rbf_add_nondir(dentry, inode);
}

static int rbf_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct super_block *s = dir->i_sb;
	struct inode *inode;
	int err = -EMLINK;

	if (dir->i_nlink >= RBF_LINK_MAX)
		goto out;

	rbf_inc_count(dir);

	inode = rbf_new_inode(dir, S_IFDIR | mode);
	err = PTR_ERR(inode);
	if (IS_ERR(inode))
		goto out_dir;

	inode->i_op = s_su_operations->dir_inode_ops;
	inode->i_fop = s_su_operations->dir_operations;
	inode->i_mapping->a_ops = s_su_operations->address_space_ops;

	rbf_inc_count(inode);

	err = rbf_make_empty(inode, dir);
	if (err)
		goto out_fail;

	err = rbf_add_link(dentry, inode);
	if (err)
		goto out_fail;

	d_instantiate(dentry, inode);
out:
	return err;

out_fail:
	rbf_dec_count(inode);
	rbf_dec_count(inode);
	iput(inode);
out_dir:
	rbf_dec_count(dir);
	goto out;
}

static int rbf_unlink(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct rbf_dirent *de;
	struct page *page;
	int err = -ENOENT;

	lock_kernel();
	de = rbf_find_entry(dir, dentry, &page);
	if (!de)
		goto out;

	err = rbf_delete_entry(dir, de, page);
	if (err)
		goto out;

	inode->i_ctime = dir->i_ctime;
	rbf_dec_count(inode);
	err = 0;
out:
	unlock_kernel();
	return err;
}

static int rbf_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	int err = -ENOTEMPTY;

	lock_kernel();
	if (rbf_empty_dir(inode)) {
		err = rbf_unlink(dir, dentry);
		if (!err) {
			inode->i_size = 0;
			rbf_dec_count(inode);
			rbf_dec_count(dir);
		}
	}
	unlock_kernel();
	return err;
}

static int rbf_rename(struct inode *old_dir, struct dentry *old_dentry,
		      struct inode *new_dir, struct dentry *new_dentry )
{
	struct inode *old_inode = old_dentry->d_inode;
	struct inode *new_inode = new_dentry->d_inode;
	struct page *dir_page = NULL;
	struct rbf_dirent *dir_de = NULL;
	struct page *old_page;
	struct rbf_dirent *old_de;
	int err = -ENOENT;

	old_de = rbf_find_entry(old_dir, old_dentry, &old_page);
	if (!old_de)
		goto out;

	if (S_ISDIR(old_inode->i_mode)) {
		err = -EIO;
		dir_de = rbf_dotdot(old_inode, &dir_page);
		if (!dir_de)
			goto out_old;
	}

	if (new_inode) {
		struct page *new_page;
		struct rbf_dirent *new_de;

		err = -ENOTEMPTY;
		if (dir_de && !rbf_empty_dir(new_inode))
			goto out_dir;

		err = -ENOENT;
		new_de = rbf_find_entry(new_dir, new_dentry, &new_page);
		if (!new_de)
			goto out_dir;
		rbf_inc_count(old_inode);
		rbf_set_link(new_dir, new_de, new_page, old_inode);
		new_inode->i_ctime = CURRENT_TIME;
		if (dir_de)
			inode_dec_link_count(new_inode);
		rbf_dec_count(new_inode);
	} else {
		if (dir_de) {
			err = -EMLINK;
			if (new_dir->i_nlink >= RBF_LINK_MAX)
				goto out_dir;
		}
		rbf_inc_count(old_inode);
		err = rbf_add_link(new_dentry, old_inode);
		if (err) {
			rbf_dec_count(old_inode);
			goto out_dir;
		}
		if (dir_de)
			rbf_inc_count(new_dir);
	}

	rbf_delete_entry(old_dir, old_de, old_page);
	rbf_dec_count(old_inode);

	if (dir_de) {
		rbf_set_link(old_inode, dir_de, dir_page, new_dir);
		rbf_dec_count(old_dir);
	}
	return 0;

out_dir:
	if (dir_de)
		rbf_release_entry(dir_page);
	
out_old:
	rbf_release_entry(old_page);
out:
	return err;
}

struct inode_operations rbf_page_dir_inode_ops = {
	.create =	rbf_create,
	.lookup =	rbf_lookup,
	.link =		rbf_link,
	.unlink =	rbf_unlink,
	.mkdir =	rbf_mkdir,
	.rmdir =	rbf_rmdir,
	.rename =	rbf_rename
};
