/*
 *	include/linux/rbf_fs.h - RBF data structures.
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 */

#ifndef _LINUX_RBF_FS_H
#define _LINUX_RBF_FS_H

/* RBF inode layout on disk */
struct rbf_inode {
	__u8	fd_att;			/* file attributes */
	__u8 	fd_own[2];		/* group:user */
	__u8	fd_date[5];		/* data last modified */
	__u8	fd_link;		/* link count */
	__u8	fd_fsize[4];		/* file size */
	__u8	fd_dcr[3];		/* data created */
};

/* file attributes */
#define	RBF_S_IREAD	0x01            /* owner read */
#define	RBF_S_IWRITE	0x02            /* owner write */
#define	RBF_S_IEXEC	0x04            /* owner execute */
#define	RBF_S_IOREAD	0x08            /* public read */
#define	RBF_S_IOWRITE	0x10            /* public write */
#define	RBF_S_IOEXEC	0x20            /* public execute */
#define	RBF_S_ISHARE	0x40            /* non-sharable */
#define	RBF_S_IFDIR	0x80            /* directory */

/* max link count */
#define RBF_LINK_MAX	255

/* segment list follows rbf_inode on disk */
struct rbf_segment {
	__u8	sg_addr[3];		/* segment begin lba */
	__u8	sg_size[2];		/* segment size in blocks */
};

#define RBF_NAMELEN		28	
#define RBF_DIRENT_SIZE		32

/* RBF directory entry on disk */
struct rbf_dirent {
	char	dir_name[RBF_NAMELEN];	/* directory name (b7 terminated) */
	__u8	dir_res1;		/* must be zero */
	__u8	dir_addr[3];		/* file descriptor address */
};

#define RBF_MAGIC	0x4372757a	/* 'Cruz' */
#define RBF_BSIZE	256		/* default block size */
#define RBF_BSIZE_BITS	8

/* RBF superblock layout on disk */
struct rbf_super_block {
	__u8	dd_tot[3];		/* total number of blocks on media */
	__u8	dd_tks;			/* track size in blocks */
	__u16	dd_map;			/* number of bytes in allocation map */
	__u16	dd_bit;			/* disk cluster size (blocks/bit) */
	__u8	dd_dir[3];		/* lba of root directory */
	__u8	dd_own[2];		/* disk owner ID */
	__u8	dd_att;			/* disk attributes */
	__u16	dd_dsk;			/* disk ID */
	__u8	dd_fmt;			/* disk format (density/sides) */
	__u8	dd_spt[2];		/* blocks/track */
	__u8	dd_res[2];		/* reserved */
	__u8	dd_bt[3];		/* system bootstrap lba or bootstrap fd lba */
	__u16	dd_bsz;			/* system bootstrap size (0 => bootstrap via fd) */
	__u8	dd_date[5];		/* disk creation date */
	__u8	dd_name[32];		/* disk name */
	__u8	dd_opt[32];		/* PD options */
	__u8	dd_res2;		/* reserved */
	__u32	dd_magic;		/* media integrity code */
	__u32	dd_maplsn;		/* bitmap starting lba (0 => lba 1) */
	__u16	dd_lsnsize;		/* logical block size (0 => 256 bytes)	*/
	__u16	dd_versid;		/* block 0 version ID */
	__u8	dd_res3[4];		/* reserved */
	__u8	dd_parttab[3];		/* partition table lba */
	__u8	dd_npart;		/* number of partitions */
	__u8	dd_res4[12];		/* reserved */
	__u8	dd_res5[128];		/* reserved */
};

/* dd_fmt (floppy disk format) */
#define	RBF_FMT_DS	(1<<0)		/* double-sided			*/
#define	RBF_FMT_DD	(1<<1)		/* double-density		*/
#define	RBF_FMT_DT	(1<<2)		/* double-track (96/135 tpi)	*/
#define	RBF_FMT_QT	(1<<3)		/* quad-track (192 tpi)		*/
#define	RBF_FMT_OT	(1<<4)		/* octal-track (384 tpi)	*/

/* RBF partition table entry */
struct rbf_partent {
	__u8	pt_id;			/* id of partition 1..n (0 => unallocated) */;
	__u8	pt_offset[3];		/* offset of partition in blocks */
	__u8	pt_size[3];		/* size of partition in blocks */
	__u8	pt_res[9];		/* reserved */
};
#define RBF_MAXPART 16

#ifdef __KERNEL__

/*
 *	read/write misaligned big-endian data fields
 */
static inline u32 rbf_read(const u8 *p, int n)
{
	u32 acc = *p++;
	
	while (--n)
		acc = (acc << 8) | *p++;
	return acc;
}

static inline void rbf_write(u8 *p, int n, u32 v)
{
	do {
		p[--n] = v;
		v >>= 8;
	}
	while (n);
}

#endif

#endif /* _LINUX_RBF_FS_H */
